@@ -144,7 +144,9 @@ class StrType(object):
         emitter.emit("}")
 
     def encode(self, emitter, value):
-        emitter.emit("if (!amqp_encode_%d(encoded, &offset, %s.len)" % (self.lenbits, value))
+        emitter.emit("if (UINT%d_MAX < %s.len" % (self.lenbits, value))
+        emitter.emit("    || !amqp_encode_%d(encoded, &offset, (uint%d_t)%s.len)" %
+                (self.lenbits, self.lenbits, value))
         emitter.emit("    || !amqp_encode_bytes(encoded, &offset, %s))" % (value,))
         emitter.emit("  return AMQP_STATUS_BAD_AMQP_DATA;")
 
@@ -225,7 +227,6 @@ apiMethodInfo  = {
     "amqp_channel_open": ["out_of_band"],
     "amqp_channel_close": False, # needs special handling
     "amqp_access_request": False, # huh?
-    "amqp_exchange_declare": ["auto_delete", "internal"],
     "amqp_basic_get": False, # get-ok has content
 }
 
@@ -285,12 +286,9 @@ def genErl(spec):
 
     def genDecodeMethodFields(m):
         print "    case %s: {" % (m.defName(),)
-        if m.arguments:
-            print "      %s *m = (%s *) amqp_pool_alloc(pool, sizeof(%s));" % \
-                (m.structName(), m.structName(), m.structName())
-            print "      if (m == NULL) { return AMQP_STATUS_NO_MEMORY; }"
-        else:
-            print "      %s *m = NULL; /* no fields */" % (m.structName(),)
+        print "      %s *m = (%s *) amqp_pool_alloc(pool, sizeof(%s));" % \
+            (m.structName(), m.structName(), m.structName())
+        print "      if (m == NULL) { return AMQP_STATUS_NO_MEMORY; }"
 
         emitter = BitDecoder(Emitter("      "))
         for f in m.arguments:
@@ -327,7 +325,7 @@ def genErl(spec):
             typeFor(spec, f).encode(emitter, "m->"+c_ize(f.name))
         emitter.flush()
 
-        print "      return offset;"
+        print "      return (int)offset;"
         print "    }"
 
     def genEncodeProperties(c):
@@ -341,7 +339,7 @@ def genErl(spec):
             typeFor(spec, f).encode(emitter, "p->"+c_ize(f.name))
             emitter.emit("}")
 
-        print "      return offset;"
+        print "      return (int)offset;"
         print "    }"
 
     methods = spec.allMethods()
