@@ -7,9 +7,8 @@ build_autotools() {
 }
 
 build_cmake() {
-  CFLAGS="-fsanitize=undefined"
   mkdir $PWD/_build && cd $PWD/_build
-  cmake .. -DCMAKE_INSTALL_PREFIX=$PWD/../_install
+  cmake .. -DCMAKE_INSTALL_PREFIX=$PWD/../_install -DCMAKE_C_FLAGS="-Werror"
   cmake --build . --target install
   ctest -V .
 }
@@ -17,7 +16,7 @@ build_cmake() {
 build_asan() {
   mkdir $PWD/_build && cd $PWD/_build
   cmake .. -DCMAKE_BUILD_TYPE=Debug -DCMAKE_INSTALL_PREFIX=$PWD/../_install \
-    -DCMAKE_C_FLAGS="-fsanitize=address,undefined -O1"
+    -DCMAKE_C_FLAGS="-Werror -fsanitize=address,undefined -O1"
   cmake --build . --target install
   ctest -V .
 }
@@ -25,17 +24,33 @@ build_asan() {
 build_tsan() {
   mkdir $PWD/_build && cd $PWD/_build
   cmake .. -DCMAKE_BUILD_TYPE=Debug -DCMAKE_INSTALL_PREFIX=$PWD/../_install \
-    -DCMAKE_C_FLAGS="-fsanitize=thread,undefined -O1"
+    -DCMAKE_C_FLAGS="-Werror -fsanitize=thread,undefined -O1"
   cmake --build . --target install
   ctest -V .
 }
 
+build_scan-build() {
+  mkdir $PWD/_build && cd $PWD/_build
+  scan-build-3.7 cmake .. -DCMAKE_BUILD_TYPE=Debug \
+    -DCMAKE_INSTALL_PREFIX=$PWD/../_install \
+    -DCMAKE_C_FLAGS="-Werror"
+  scan-build-3.7 make install
+}
+
 if [ "$#" -ne 1 ]; then
-  echo "Usage: $0 {autotools|cmake|asan|tsan}"
+  echo "Usage: $0 {autotools|cmake|asan|tsan|scan-build}"
   exit 1
 fi
 
 set -e  # exit on error.
 set -x  # echo commands.
 
+case $TRAVIS_OS_NAME in
+osx)
+  # This prints out a long list of updated packages, which isn't useful.
+  brew update > /dev/null
+  brew install popt
+  ;;
+esac
+
 eval "build_$1"
